(declaim (optimize (speed 1) (safety 3) (debug 3)))

(in-package :tim-judge)

;; Change this according to your development directory
(defparameter *dev-dir*  #p"/mnt/Workspace/Projects/src/tim-judge/trunk/")

(defparameter *bash*                   "/bin/bash")
(defparameter *output-archive-dir*     "output/")

(defparameter *text-compare-helper*    "compare-text.sh")
(defparameter *run-a-test-helper*      "run-test.sh")
(defparameter *cleanup-all-program*    "cleanup-all.sh")

(defparameter *problem-set*           nil)
(defparameter *solution-set*          nil)
(defparameter *test-set-set*         nil)

(defparameter *working-dir* (namestring (osicat:current-directory)))

(defparameter *problem-set-mutex*   (make-mutex :name "problem-set"))
(defparameter *solution-set-mutex*  (make-mutex :name "solution-set"))
(defparameter *test-set-set-mutex* (make-mutex :name "test-set-set"))

(defun add-problem (problem &optional (problem-set *problem-set*))
  (when (null (problem-code->problem (conf-problem-code problem)))
    (with-mutex (*problem-set-mutex*)
      (push problem problem-set))))

(defun add-problem-details (code name input-dir
                            &optional (judge-program *text-compare-helper*)
                            (time-limit 2.0)
                            (mem-limit (* 2.0 1024 1024))
                            (problem-set *problem-set*))
  (add-problem (make-instance 'tim-problem
                              :problem-code code
                              :problem-name name
                              :input-dir input-dir
                              :judge-program judge-program
                              :time-limit time-limit
                              :mem-limit mem-limit)
               problem-set))

(defun add-solution (solution &optional (solution-set *solution-set*))
  (when (null (solution-id->solution (conf-solution-id solution)))
    (with-mutex (*solution-set-mutex*)
      (push solution solution-set))))

(defun add-solution-details (problem-code solution-file language
                              solution-id
                              &optional (solution-output nil))
  (if (null solution-output)
      (setf solution-output (format nil "~A" solution-id)))
  (add-solution (make-instance 'tim-solution
                               :problem-code problem-code
                               :solution-file solution-file
                               :solution-output solution-output
                               :language language
                               :solution-id solution-id)))

(defun test-type->extension (type)
  (or (getf *test-extensions* type) ".txt"))

(defun language->compiler (lang)
  (or (getf *languages-compilers* lang) ""))

(defun language->interpreter (lang)
  (or (getf *languages-interpreters* lang) ""))

(defun language->extension (lang)
  (or (getf *languages-extensions* lang) ""))

(defun replace-extension (file-name old-ext new-ext)
  "Replace a file extension."
  (concatenate 'string
               (subseq file-name 0 (search old-ext file-name :from-end t))
               new-ext))

(defun cleanup-all ()
  "Clean-up all temporary files."
  (run-program *bash*
               (list *cleanup-all-program*)
               :wait t :output nil))

(defun make-compile-helper (language &optional
                            (working-dir *working-dir*))
  "Generating the helper corresponding to the language."
  (concatenate 'string working-dir "compile-"
               (language->compiler language) ".sh"))

(defun make-compile-command (solution &optional
                             (working-dir *working-dir*))
  (list (make-compile-helper (conf-language solution)
                             working-dir)
        (conf-solution-file solution)
        (conf-solution-output solution)))

(defun compile-solution (solution &optional
                         (working-dir *working-dir*))
  "Compile a source file."
  (let* ((out-stream (make-string-output-stream))
         (err-stream (make-string-output-stream))
         (p (run-program *bash*
                         (make-compile-command solution working-dir)
                         :wait t :input nil :pty nil
                         :output out-stream
                         :error err-stream)))
    (values (process-exit-code p)
            (get-output-stream-string out-stream)
            (get-output-stream-string err-stream))))

(defun compare-text (file-1 file-2
                     &key (helper *text-compare-helper*))
  "Comparing file-1 and file-2 using a helper and text comparison method."
  (process-exit-code (run-program *bash*
                                  (list helper
                                        (namestring file-1)
                                        (namestring file-2))
                                  :wait t :input nil :output nil)))

(defun make-test-file (test-type number)
  "Make an input/output/answer file for testing."
  (if (stringp number)
      (concatenate 'string number "." (test-type->extension test-type))
      (concatenate 'string (write-to-string number) "."
                   (test-type->extension test-type))))

(defun problem-code->problem (code &optional (problem-set *problem-set*))
  "Get the problem based on its code.  Returns nil if not found."
  (find-if #'(lambda (x)
               (equal code (conf-problem-code x)))
           problem-set))

(defun solution-id->solution (id &optional (solution-set *solution-set*))
  "Get the solution based on its id.  Returns nil if not found."
  (find-if #'(lambda (x)
               (equal id (conf-solution-id x)))
           solution-set))

(defun make-test-output (solution-id test-num)
  (format nil "~A~A_~A.~A"
          *output-archive-dir*
          solution-id test-num
          (test-type->extension :output)))

(defun make-run-test-command (problem solution
                              &optional (test-num 1)
                              (helper *run-a-test-helper*))
  (list helper
        (conf-solution-output solution)
        (format nil "~A" test-num)
        (conf-input-dir problem)
        (make-test-output (conf-solution-id solution)
                          test-num)
        (language->interpreter (conf-language solution))))

(defun run-test (problem solution &optional (test-num 1))
  "Run a specific test."
  (let* ((out-stream (make-string-output-stream))
         (err-stream (make-string-output-stream))
         (command (make-run-test-command problem
                                         solution
                                         test-num))
         (p (run-program *bash*
                         command
                         :output out-stream
                         :error err-stream
                         :wait t)))
    (values (process-exit-code p)
            (get-output-stream-string out-stream)
            (get-output-stream-string err-stream))))

(defun make-result-form (test-num result)
  "Make the result form: '(test-num result)."
  (list test-num result))

(defun make-answer-file (dir test-num)
  "Construct the answer for a test."
  (concatenate 'string dir (make-test-file :answer test-num)))

(defun judge-text-compare (file-1 file-2
                           &key (helper *text-compare-helper*))
  "Convert exit-code to judge points."
  (- 1 (compare-text file-1 file-2 :helper helper)))

(defun judge-result (problem solution test-num
                     &optional (checking-method (function judge-text-compare)))
  "With the helper, check the result and the answer."
  (let* ((output-file (make-test-output (conf-solution-id solution)
                                        test-num))
         (answer-file (make-answer-file (conf-input-dir problem)
                                        test-num)))
    (funcall checking-method answer-file output-file)))

(defun run-test-set (test-set &key (solution-set *solution-set*)
                     (problem-set *problem-set*))
  "Run a test-set and collection the results.  If there is something
wrong while testing, it stops immediately"
  (with-accessors ((from-num conf-from-test) (to-num conf-to-test)
                   (problem-code conf-problem-code)
                   (solution-id conf-solution-id))
      test-set
    ;; Debug code
    ;; (format t "Problem Code: ~A; Solution Id: ~A~%" problem-code solution-id)
    ;; 
    (let* ((problem (problem-code->problem problem-code problem-set))
           (solution (solution-id->solution solution-id solution-set))
           (result-list nil))
      (loop
         for test-num from from-num to to-num
         do
           (progn
             (let* ((exit-code (run-test problem solution test-num)))
               (if (zerop exit-code)
                   (push (make-result-form test-num
                                           (judge-result problem
                                                         solution test-num))
                         result-list)
                   (progn
                     (setf result-list exit-code)
                     (return))))
             ;; (multiple-value-bind (exit-code output error)
             ;;     (run-test problem solution test-num)
             ;;   (if (zerop exit-code)
             ;;       (push (make-result-form test-num
             ;;                               (judge-result problem
             ;;                                             solution test-num))
             ;;             result-list)
             ;;       (progn
             ;;         (setf result-list exit-code)
             ;;         (return))))
             ))
      result-list)))

(defun print-test-set-result (results &key (stream *standard-output*))
  "Print the result for each test in a test set."
  (format stream "~{~{~5d ~a~}~%~}" results))

(defun set-development-environment ()
  (setf *default-pathname-defaults* *dev-dir*)
  (setf (osicat:current-directory) *dev-dir*))

(defun initialize (&key
                   (conf-data-structures "conf-data-structures.lisp")
                   (conf-problems  "conf-problems.lisp")
                   (conf-solutions "conf-solutions.lisp")
                   (conf-network   "conf-network.lisp"))
  "Initialize the system, run only once."
  (load (compile-file conf-data-structures))
  (make-class-from-structure *problem-class-structure*   'tim-problem)
  (make-class-from-structure *solution-class-structure*  'tim-solution)
  (make-class-from-structure *test-set-class-structure* 'tim-test-set)
  (load (compile-file conf-problems))
  (load (compile-file conf-solutions))
  (load (compile-file conf-network)))
